mutable struct InPlaceFlashBuffer
    z
    forces
    function InPlaceFlashBuffer(n)
        z = zeros(n)
        new(z, nothing)
    end
end

struct FlashResults <: ScalarVariable
    storage
    method
    update_buffer
    use_threads::Bool
    function FlashResults(system; method = SSIFlash(), threads = Threads.nthreads() > 1)
        eos = system.equation_of_state
        nc = MultiComponentFlash.number_of_components(eos)
        if has_other_phase(system)
            n = nc + 1
        else
            n = nc
        end
        storage = []
        buffers = []
        if threads
            N = Threads.nthreads()
        else
            N = 1
        end
        for i = 1:N
            s = flash_storage(eos, method = method, inc_jac = true, diff_externals = true, npartials = n, static_size = true)
            push!(storage, s)
            b = InPlaceFlashBuffer(nc)
            push!(buffers, b)
        end
        storage = tuple(storage...)
        buffers = tuple(buffers...)
        new(storage, method, buffers, threads)
    end
end

default_value(model, ::FlashResults) = FlashedMixture2Phase(model.system.equation_of_state)

function initialize_variable_value(model, pvar::FlashResults, val::AbstractDict; need_value = false)
    @assert need_value == false
    n = number_of_entities(model, pvar)
    v = default_value(model, pvar)
    T = typeof(v)
    V = Vector{T}()
    sizehint!(V, n)
    for i in 1:n
        push!(V, default_value(model, pvar))
    end
    initialize_variable_value(model, pvar, V)
end

function Jutul.initialize_variable_value!(state, model, pvar::FlashResults, symb, val::AbstractDict; kwarg...)
    state[symb] = initialize_variable_value(model, pvar, val; kwarg...)
end

function initialize_variable_ad!(state, model, pvar::FlashResults, symb, npartials, diag_pos; context = DefaultContext(), kwarg...)
    n = number_of_entities(model, pvar)
    v_ad = get_ad_entity_scalar(1.0, npartials, diag_pos; kwarg...)
    ∂T = typeof(v_ad)
    eos = model.system.equation_of_state

    r = FlashedMixture2Phase(eos, ∂T)
    T = typeof(r)
    V = Vector{T}(undef, n)
    for i in 1:n
        V[i] = FlashedMixture2Phase(eos, ∂T)
    end
    state[symb] = V
    return state
end

@jutul_secondary function update_flash!(flash_results, fr::FlashResults, model, Pressure, Temperature, OverallMoleFractions, ix)
    flash_entity_loop!(flash_results, fr, model, Pressure, Temperature, OverallMoleFractions, nothing, ix)
end

@jutul_secondary function update_flash!(flash_results, fr::FlashResults, model::LVCompositionalModel3Phase, Pressure, Temperature, OverallMoleFractions, ImmiscibleSaturation, ix)
    flash_entity_loop!(flash_results, fr, model, Pressure, Temperature, OverallMoleFractions, ImmiscibleSaturation, ix)
end

function flash_entity_loop!(flash_results, fr, model, Pressure, Temperature, OverallMoleFractions, sw, ix)
    storage, m, buffers = fr.storage, fr.method, fr.update_buffer
    eos = model.system.equation_of_state

    S, buf = thread_buffers(storage, buffers)
    update_flash_buffer!(buf, eos, Pressure, Temperature, OverallMoleFractions)
    buf_z = buf.z
    buf_forces = buf.forces
    @inbounds for i in ix
        flash_results[i] = internal_flash!(flash_results[i], S, m, eos, buf_z, buf_forces, Pressure, Temperature, OverallMoleFractions, sw, i)
    end
end

@inline function Jutul.update_values!(vals::AbstractVector{<:FlashedMixture2Phase}, next::AbstractVector{<:FlashedMixture2Phase})
    function replace_flashed_phase_values(x, y)
        Jutul.update_values!(x.mole_fractions, y.mole_fractions)
        Z = x.Z
        return FlashedPhase(x.mole_fractions, Z - value(Z) + value(y.Z))
    end

    for i in eachindex(vals)
        old_v = vals[i]
        next_v = next[i]
        l = replace_flashed_phase_values(old_v.liquid, next_v.liquid)
        v = replace_flashed_phase_values(old_v.vapor, next_v.vapor)
        K = old_v.K
        V = old_v.V
        Jutul.update_values!(K, next_v.K)
        state = next_v.state
        V = V - value(V) + value(next_v.V)
        vals[i] = FlashedMixture2Phase(state, K, V, l, v)
    end
end

function thread_buffers(storage, buffers)
    thread_id = Threads.threadid()
    S = storage[thread_id]
    thread_buffer = buffers[thread_id]
    return (S, thread_buffer)
end

function update_flash_buffer!(buf, eos, Pressure, Temperature, OverallMoleFractions)
    if isnothing(buf.forces) || eltype(buf.forces.A_ij) != eltype(OverallMoleFractions)
        P = Pressure[1]
        T = Temperature[1]
        if isa(OverallMoleFractions, AbstractVector)
            Z = OverallMoleFractions
        else
            Z = OverallMoleFractions[:, 1]
        end
        buf.forces = force_coefficients(eos, (p = P, T = T, z = Z), static_size = true)
    end
end

function internal_flash!(f, S, m, eos, buf_z, buf_forces, Pressure, Temperature, OverallMoleFractions, sw, i)
    @inline function immiscible_sat(::Nothing, i)
        return 0.0
    end

    @inline function immiscible_sat(s, i)
        return @inbounds s[i]
    end

    @inbounds begin
        P = Pressure[i]
        T = Temperature[i]
        Z = @view OverallMoleFractions[:, i]
        Sw = immiscible_sat(sw, i)

        K = f.K
        x = f.liquid.mole_fractions
        y = f.vapor.mole_fractions

        return update_flash_result(S, m, eos, K, x, y, buf_z, buf_forces, P, T, Z, Sw)
    end
end


function update_flash_result(S, m, eos, K, x, y, z, forces, P, T, Z, Sw = 0.0)
    @. z = max(value(Z), 1e-8)
    # Conditions
    c = (p = value(P), T = value(T), z = z)
    # Perform flash
    if is_pure_single_phase(Sw)
        vapor_frac = NaN
    else
        vapor_frac = flash_2ph!(S, K, eos, c, NaN, method = m, extra_out = false, z_min = nothing)
    end
    force_coefficients!(forces, eos, (p = P, T = T, z = Z))
    if isnan(vapor_frac)
        # Single phase condition. Life is easy.
        Z_L, Z_V, V, phase_state = single_phase_update!(P, T, Z, x, y, forces, eos, c)
    else
        # Two-phase condition: We have some work to do.
        Z_L, Z_V, V, phase_state = two_phase_update!(S, P, T, Z, x, y, K, vapor_frac, forces, eos, c)
    end
    out = FlashedMixture2Phase(phase_state, K, V, x, y, Z_L, Z_V)
    return out
end

function get_compressibility_factor(forces, eos, P, T, Z, phase = :unknown)
    ∂cond = (p = P, T = T, z = Z)
    force_coefficients!(forces, eos, ∂cond)
    scalars = force_scalars(eos, ∂cond, forces)
    return mixture_compressibility_factor(eos, ∂cond, forces, scalars, phase)
end

@inline function single_phase_update!(P, T, Z, x, y, forces, eos, c)
    AD = Base.promote_type(eltype(Z), typeof(P), typeof(T))
    Z_L = get_compressibility_factor(forces, eos, P, T, Z)
    Z_V = Z_L
    @. x = Z
    @. y = Z
    V = single_phase_label(eos, c)
    if V > 0.5
        phase_state = MultiComponentFlash.single_phase_v
    else
        phase_state = MultiComponentFlash.single_phase_l
    end
    V = convert(AD, V)
    out = (Z_L::AD, Z_V::AD, V::AD, phase_state::PhaseState2Phase)
    return out
end

function two_phase_pre!(S, P, T, Z, x::AbstractVector{AD}, y::AbstractVector{AD}, vapor_frac, eos, c) where {AD <: ForwardDiff.Dual}
    V = convert(AD, vapor_frac)
    if P isa AD || T isa AD || eltype(Z)<:AD
        inverse_flash_update!(S, eos, c, vapor_frac)
        ∂c = (p = P, T = T, z = Z)
        V = set_partials_vapor_fraction(V, S, eos, ∂c)
        set_partials_phase_mole_fractions!(x, S, eos, ∂c, :liquid)
        set_partials_phase_mole_fractions!(y, S, eos, ∂c, :vapor)
    end
    return V
end

two_phase_pre!(S, P, T, Z, x, y, V, eos, c) = V


@inline function two_phase_update!(S, P, T, Z, x, y, K, vapor_frac, forces, eos, c)
    AD = Base.promote_type(typeof(P), eltype(Z), typeof(T), eltype(x), eltype(y))
    @. x = liquid_mole_fraction(Z, K, vapor_frac)
    @. y = vapor_mole_fraction(x, K)
    V = two_phase_pre!(S, P, T, Z, x, y, vapor_frac, eos, c)
    Z_L = get_compressibility_factor(forces, eos, P, T, x, :liquid)
    Z_V = get_compressibility_factor(forces, eos, P, T, y, :vapor)
    phase_state = MultiComponentFlash.two_phase_lv

    return (convert(AD, Z_L), convert(AD, Z_V), convert(AD, V), phase_state)
end
